"""
Source: http://code.activestate.com/recipes/577781-pluralize-word-convert-singular-word-to-its-plural/
"""

ABERRANT_PLURAL_MAP = {
    'appendix': 'appendices',
    'barracks': 'barracks',
    'cactus': 'cacti',
    'child': 'children',
    'criterion': 'criteria',
    'deer': 'deer',
    'echo': 'echoes',
    'elf': 'elves',
    'embargo': 'embargoes',
    'focus': 'foci',
    'fungus': 'fungi',
    'goose': 'geese',
    'hero': 'heroes',
    'hoof': 'hooves',
    'index': 'indices',
    'knife': 'knives',
    'leaf': 'leaves',
    'life': 'lives',
    'man': 'men',
    'mouse': 'mice',
    'nucleus': 'nuclei',
    'person': 'people',
    'phenomenon': 'phenomena',
    'potato': 'potatoes',
    'self': 'selves',
    'syllabus': 'syllabi',
    'tomato': 'tomatoes',
    'torpedo': 'torpedoes',
    'veto': 'vetoes',
    'woman': 'women',
}

VOWELS = set('aeiou')


def pluralize(singular):
    """Return plural form of given lowercase singular word (English only). Based on
    ActiveState recipe http://code.activestate.com/recipes/413172/

    >>> pluralize('')
    ''
    >>> pluralize('goose')
    'geese'
    >>> pluralize('dolly')
    'dollies'
    >>> pluralize('genius')
    'genii'
    >>> pluralize('jones')
    'joneses'
    >>> pluralize('pass')
    'passes'
    >>> pluralize('zero')
    'zeros'
    >>> pluralize('casino')
    'casinos'
    >>> pluralize('hero')
    'heroes'
    >>> pluralize('church')
    'churches'
    >>> pluralize('x')
    'xs'
    >>> pluralize('car')
    'cars'

    """
    if not singular:
        return ''
    plural = ABERRANT_PLURAL_MAP.get(singular)
    if plural:
        return plural
    root = singular
    try:
        if singular[-1] == 'y' and singular[-2] not in VOWELS:
            root = singular[:-1]
            suffix = 'ies'
        elif singular[-1] == 's':
            if singular[-2] in VOWELS:
                if singular[-3:] == 'ius':
                    root = singular[:-2]
                    suffix = 'i'
                else:
                    root = singular[:-1]
                    suffix = 'ses'
            else:
                suffix = 'es'
        elif singular[-2:] in ('ch', 'sh'):
            suffix = 'es'
        else:
            suffix = 's'
    except IndexError:
        suffix = 's'
    plural = root + suffix
    return plural



from .find_edges_connected_to_vertex import find_edges_connected_to_vertex
from .is_edge_parallel_to_axis import (is_edge_parallel_to_x_axis,
                                       is_edge_parallel_to_y_axis,
                                       is_edge_parallel_to_z_axis)



from collections import OrderedDict
from itertools import groupby
from operator import itemgetter

from .pluralize import pluralize


def merge_cut_list_items_by_length(cut_list_rows):
    """Merge cut list items with same length.

    This will combine descriptions.

    Note, we also pluralize the 'Description' if needed,
    so rows should have the 'Description' in singular form.

    :param cut_list_rows: cut list rows
    :type cut_list_rows: list(dict)
    """
    merged_cut_list_rows = []
    rows_by_length = group_by_length(cut_list_rows)
    for length, rows in rows_by_length:
        row_list = list(rows)
        first_row = row_list[0]
        if len(row_list) == 1:
            merged_cut_list_rows.append(OrderedDict([
                ('Quantity', first_row['Quantity']),
                ('Description', get_potentially_pluralized_description(first_row)),
                ('Length', first_row['Length'])
            ]))
        else:  # multiple rows with same length
            merged_cut_list_rows.append(OrderedDict([
                ('Quantity', sum_quantities(row_list)),
                ('Description', concatenate_descriptions(row_list)),
                ('Length', first_row['Length'])
            ]))
    return merged_cut_list_rows


def group_by_length(cut_list_rows):
    rows_sorted_by_length = sorted(cut_list_rows, key=lambda r: r['Length'])
    return groupby(rows_sorted_by_length, key=itemgetter('Length'))


def sum_quantities(cut_list_rows):
    return str(sum([int(row['Quantity']) for row in cut_list_rows]))


def concatenate_descriptions(cut_list_rows):
    first_row = cut_list_rows[0]
    description = stringify_cut_list_row(first_row)
    if len(cut_list_rows) == 2:
        second_row = cut_list_rows[1]
        return description + ' & ' + stringify_cut_list_row(second_row)

    rows_excluding_first_and_last = cut_list_rows[1:-1]
    for row in rows_excluding_first_and_last:
        description += ', ' + stringify_cut_list_row(row)

    last_row = cut_list_rows[len(cut_list_rows) - 1]
    return description + ', & ' + stringify_cut_list_row(last_row)


def stringify_cut_list_row(cut_list_row):
    return '({}) {}'.format(
        cut_list_row['Quantity'],
        get_potentially_pluralized_description(cut_list_row)
    )


def get_potentially_pluralized_description(cut_list_row):
    description = cut_list_row['Description']
    if int(cut_list_row['Quantity']) <= 1:
        return description
    else:
        return pluralize(description)


def pluralize_last_word(phrase):
    words = phrase.split(' ')
    last_word = words[len(words) - 1]
    pluralized_last_word = pluralize(last_word)
    return ' '.join(words[:-1] + [pluralized_last_word])



from FreeCAD import Vector


def is_face_parallel_to_yz_plane(face):
    x_axis = Vector(1, 0, 0)
    return _is_face_parallel_to_plane(face, x_axis)


def is_face_parallel_to_xz_plane(face):
    y_axis = Vector(0, 1, 0)
    return _is_face_parallel_to_plane(face, y_axis)


def is_face_parallel_to_xy_plane(face):
    z_axis = Vector(0, 0, 1)
    return _is_face_parallel_to_plane(face, z_axis)


def _is_face_parallel_to_plane(face, axis_vector):
    return axis_vector == Vector(
        abs(round(face.Surface.Axis.x)),
        abs(round(face.Surface.Axis.y)),
        abs(round(face.Surface.Axis.z))
    )



def is_edge_parallel_to_x_axis(edge):
    return _is_edge_parallel_to_axis(edge, 'x')


def is_edge_parallel_to_y_axis(edge):
    return _is_edge_parallel_to_axis(edge, 'y')


def is_edge_parallel_to_z_axis(edge):
    return _is_edge_parallel_to_axis(edge, 'z')


def _is_edge_parallel_to_axis(edge, axis):
    index_by_axis = {'x': 0, 'y': 1, 'z': 2}
    inverse_orientations = filter(
        lambda item: item[0] != axis, index_by_axis.items())
    a_index, b_index = map(lambda item: item[1], inverse_orientations)
    first_point = edge.valueAt(edge.FirstParameter)
    a1 = round(first_point[a_index])
    b1 = round(first_point[b_index])

    last_point = edge.valueAt(edge.LastParameter)
    a2 = round(last_point[a_index])
    b2 = round(last_point[b_index])

    return a1 == a2 and b1 == b2



from collections import OrderedDict


def convert_cut_list_to_ordered_dicts(cut_list):
    return list(map(convert_cut_list_item_to_ordered_dict, cut_list))


def convert_cut_list_item_to_ordered_dict(cut_list_dict):
    return OrderedDict([
        ('Quantity', cut_list_dict['quantity']),
        ('Description', cut_list_dict['description']),
        ('Length', cut_list_dict['length'])
    ])



class Model:
    """
    Base class for models that encapsulate the data (i.e. topography and shape)
    for a part, and is separate from the "view" or GUI representation.
    """

    def __init__(self, obj):
        """
        Constructor

        Arguments
        ---------
        - obj: Created with document.addObject('Part::FeaturePython', '{name}')
        """
        self.Object = obj

        obj.Proxy = self

    def onDocumentRestored(self, fp):
        """Executed after a document is restored,
        or a FeaturePython object is copied or duplicated.

        :param fp: Custom feature python object
        :type fp: Part::FeaturePython
        """
        self.Object = fp



from itertools import tee

import Part


def make_face_from_vectors(vectors):
    """Make a Face from list of vectors.

    See Part::TopoShapeFacePy Class Reference:
        https://www.freecadweb.org/api/d9/d35/classPart_1_1TopoShapeFacePy.html

    :raises ValueError: When there's less than three vectors in a list.
    :raises ValueError: When vectors don't form a closed wire.
    :return: A face
    :rtype: Part.Face
    """
    if len(vectors) < 3:
        raise ValueError('Must have at least 3 vectors to make a face.')
    vertices = vectors + [vectors[0]]
    edges = []
    for pair in _pair_wise(vertices):
        edge = Part.LineSegment(*pair)
        edges.append(edge)

    shape = Part.Shape(edges)
    wire = Part.Wire(shape.Edges)
    if not wire.isClosed():
        raise ValueError('Vectors must form closed face.')
    return Part.Face(wire)


def _pair_wise(iterable):
    """
    s -> (s0,s1), (s1,s2), (s2, s3), ...
    """
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)



"""Package for code relating to creating a cut list in the GUI.
"""
from .create_cut_list_task_panel import (
    create_copy_cut_list_to_clipboard_task_panel,
    create_save_cut_list_to_file_task_panel)



from .is_face_parallel_to_plane import (is_face_parallel_to_xy_plane,
                                        is_face_parallel_to_xz_plane,
                                        is_face_parallel_to_yz_plane)
from .make_face_from_vectors import make_face_from_vectors



"""A set of enumerations relating to three-dimensional space.
"""


class CoordinateAxis:
    X = 'x'
    Y = 'y'
    Z = 'z'


class Side:
    FRONT = 'front'
    TOP = 'top'
    RIGHT = 'right'
    REAR = 'rear'
    BOTTOM = 'bottom'
    LEFT = 'left'


class Plane:
    XY = 'xy'
    YZ = 'yz'
    XZ = 'xz'



"""Package for code relating to building a cut list.
"""
from .convert_cut_list_to_ordered_dicts import \
    convert_cut_list_to_ordered_dicts
from .convert_dict_list_to_wiki_list_markup import \
    convert_dict_list_to_wiki_list_markup
from .convert_dict_list_to_wiki_table_markup import \
    convert_dict_list_to_wiki_table_markup
from .merge_cut_list_items_by_length import merge_cut_list_items_by_length
from .write_dict_list_to_csv import write_dict_list_to_csv



import io
from os import path

from setuptools import setup

version = {}
with open('osecore/version.py') as fp:
    exec(fp.read(), version)

current_dir = path.abspath(path.dirname(__file__))
with io.open(path.join(current_dir, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='ose-workbench-core',
    version=version['__version__'],
    packages=[
        'osecore',
        'osecore.app',
        'osecore.app.attachment',
        'osecore.app.cut_list',
        'osecore.app.shape',
        'osecore.app.shape.edge',
        'osecore.app.shape.face',
        'osecore.gui',
        'osecore.gui.cut_list',
        'osecore.gui.selection'
    ],
    author='G Roques',
    url='https://github.com/gbroques/ose-workbench-core',
    description='Core library common to all OSE workbenches.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[],
    include_package_data=True,
    classifiers=[
        # Full List: https://pypi.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
        'Programming Language :: Python :: 3 :: Only'
    ]
)



"""Package for code relating to selection in GUI.
"""
from .selection_object import (find_edge_in_selection_object,
                               find_face_in_selection_object,
                               find_selection_object_by_type,
                               find_vertex_in_selection_object)
